import time
import logging
from wandb import wandb_dir
from wandb.pusher import AgentPuller


class Agent(object):
    def __init__(self, entity=None):
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(logging.FileHandler("agent.log", "w"))
        self.puller = AgentPuller(self, entity=entity, logger=self.logger)

    def run(self):
        self.logger.info('WandB Agent Starting')
        self.puller.run()

    def handle_event(self, event):
        pass



from git import Repo, exc
import os


class GitRepo(object):
    def __init__(self, root=None, remote="origin", lazy=True):
        self.remote_name = remote
        self._root = root
        self._repo = None
        if not lazy:
            self.repo

    @property
    def repo(self):
        if self._repo is None:
            if self.remote_name is None:
                self._repo = False
            else:
                try:
                    self._repo = Repo(self._root or os.getcwd(),
                                      search_parent_directories=True)
                except exc.InvalidGitRepositoryError:
                    self._repo = False
        return self._repo

    @property
    def enabled(self):
        return self.repo

    @property
    def root(self):
        if not self.repo:
            return False
        return self.repo.git.rev_parse("--show-toplevel")

    @property
    def dirty(self):
        if not self.repo:
            return False
        return self.repo.is_dirty()

    @property
    def last_commit(self):
        if not self.repo:
            return None
        if not self.repo.head:
            return None
        if len(self.repo.refs) > 0:
            return self.repo.head.commit.hexsha
        else:
            return self.repo.git.show_ref("--head").split(" ")[0]

    @property
    def branch(self):
        if not self.repo:
            return None
        return self.repo.head.ref.name

    @property
    def remote(self):
        if not self.repo:
            return None
        try:
            return self.repo.remotes[self.remote_name]
        except IndexError:
            return None

    @property
    def remote_url(self):
        if not self.remote:
            return None
        return self.remote.url

    def get_upstream_fork_point(self):
        """Get the most recent ancestor of HEAD that occurs on an upstream
        branch.

        First looks at the current branch's tracking branch, if applicable. If
        that doesn't work, looks at every other branch to find the most recent
        ancestor of HEAD that occurs on a tracking branch.

        Returns:
            git.Commit object or None
        """
        possible_relatives = []
        if not self.repo:
            return None
        try:
            active_branch = self.repo.active_branch
        except (TypeError, ValueError):
            return None  # detached head
        else:
            tracking_branch = active_branch.tracking_branch()
            if tracking_branch:
                possible_relatives.append(tracking_branch.commit)

        if not possible_relatives:
            for branch in self.repo.branches:
                tracking_branch = branch.tracking_branch()
                if tracking_branch is not None:
                    possible_relatives.append(tracking_branch.commit)

        head = self.repo.head
        most_recent_ancestor = None
        for possible_relative in possible_relatives:
            # at most one:
            for ancestor in self.repo.merge_base(head, possible_relative):
                if most_recent_ancestor is None:
                    most_recent_ancestor = ancestor
                elif most_recent_ancestor.is_ancestor(ancestor):
                    most_recent_ancestor = ancestor

        return most_recent_ancestor

    def tag(self, name, message):
        try:
            return self.repo.create_tag("wandb/" + name, message=message, force=True)
        except exc.GitCommandError:
            print("Failed to tag repository.")
            return None

    def push(self, name):
        if self.remote:
            try:
                return self.remote.push("wandb/" + name, force=True)
            except exc.GitCommandError:
                return None



import collections
import os
import time

import wandb
from wandb import util


class JsonlFile(object):
    """Used to store data that changes over time during runs. """

    def __init__(self, fname, out_dir='.'):
        self._start_time = time.time()
        self.fname = os.path.join(out_dir, fname)
        self._file = open(self.fname, 'w')
        self.rows = []

    def keys(self):
        if self.rows:
            return self.rows[0].keys()
        return []

    def column(self, key):
        return [r[key] for r in self.rows]

    def add(self, row):
        if not isinstance(row, collections.Mapping):
            raise wandb.Error('history.add expects dict-like')
        row['_runtime'] = time.time() - self._start_time
        self.rows.append(row)
        self._file.write(util.json_dumps_safer(row))
        self._file.write('\n')
        self._file.flush()



import copy
import multiprocessing
import os
from six.moves import queue
import signal
import sys
import time

from wandb import wandb_run
from wandb import util
from wandb import sync
from wandb import wandb_api
from wandb import streaming_log


class RunnerError(Exception):
    pass


class SingleRun(multiprocessing.Process):
    def __init__(self, api, program, args, id, dir, configs, message, show, show_output, sweep_id, **kwargs):
        super(SingleRun, self).__init__(**kwargs)
        self._api = api
        self._program = program
        self._args = args
        self._id = id
        self._dir = dir
        self._configs = configs
        self._message = message
        self._show = show
        self._result_queue = multiprocessing.Queue()
        self._show_output = show_output
        self._sweep_id = sweep_id
        self._stop_queue = multiprocessing.Queue()

    def run(self):
        # The process

        if self._id is None:
            self._id = wandb_run.generate_id()
        if self._dir is None:
            self._dir = wandb_run.run_dir_path(self._id, dry=False)
            util.mkdir_exists_ok(self._dir)
        if self._message:
            open(os.path.join(dir, 'description.md'),
                 'w').write('%s\n' % self._message)

        # setup child environment
        env = copy.copy(os.environ)
        # tell child python interpreters we accept utf-8
        # TODO(adrian): is there a language-agnostic way of doing this?
        env['PYTHONIOENCODING'] = 'UTF-8'
        env['WANDB_MODE'] = 'run'
        env['WANDB_RUN_ID'] = self._id
        env['WANDB_RUN_DIR'] = self._dir
        if self._configs:
            env['WANDB_CONFIG_PATHS'] = self._configs
        if self._show:
            env['WANDB_SHOW_RUN'] = '1'
        if self._sweep_id:
            env['WANDB_SWEEP_ID'] = self._sweep_id
        #env['WANDB_INITED'] = '1'

        # if self._api.api_key is None:
        #    raise wandb_api.Error(
        #        "No API key found, run `wandb login` or set WANDB_API_KEY")
        #run = wandb_run.Run(self._id, self._dir, None)
        # self._api.set_current_run_id(run.id)
        #syncer = sync.Sync(self._api, 'train', run, config=run.config)
        #syncer.watch(files='*', show_run=self._show)

        # stdout_stream = streaming_log.TextStreamPusher(
        #    self._api.get_file_stream_api(), 'output.log', prepend_timestamp=True)
        # stderr_stream = streaming_log.TextStreamPusher(
        #    self._api.get_file_stream_api(), 'output.log', line_prepend='ERROR',
        #    prepend_timestamp=True)

        command = [self._program] + list(self._args)
        runner = util.find_runner(self._program)
        if runner:
            command = runner.split() + command
        proc = util.SafeSubprocess(command, read_output=self._show_output)
        result = {'type': 'start', 'run_id': self._id}
        try:
            proc.run()
        except (OSError, IOError):
            result['error'] = 'Could not find program: %s' % self._program
        self._result_queue.put(result)
        if result.get('error'):
            return

        # ignore SIGINT (ctrl-c), the child process will handle, and we'll
        # exit when the child process does.
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        exitcode = None
        while True:
            time.sleep(0.1)
            try:
                self._stop_queue.get_nowait()
                print('Sending SIGTERM')
                proc.terminate()
            except queue.Empty:
                pass
            exitcode, stdout, stderr = proc.poll()
            for line in stdout:
                # stdout_stream.write(line)
                if self._show_output:
                    sys.stdout.write(line)
            for line in stderr:
                # stderr_stream.write(line)
                if self._show_output:
                    sys.stderr.write(line)
            if exitcode is not None:
                break
        # stdout_stream.close()
        # stderr_stream.close()
        # syncer.stop()

        self._result_queue.put({'type': 'finish', 'exitcode': exitcode})

    def next_event(self):
        return self._result_queue.get()

    def is_running(self):
        try:
            finish_event = self._result_queue.get_nowait()
            return False
        except queue.Empty:
            return True

    def launch(self):
        self.start()

    def term(self):
        self._stop_queue.put(True)


class Runner(object):
    def __init__(self, api):
        self._api = api
        self._runs = {}

    def run(self, program, args, id=None, dir=None,
            configs=None, message=None, show=None, show_output=True, sweep_id=None):
        print('Run: %s %s' % (program, args))
        run = SingleRun(self._api, program, args, id,
                        dir, configs, message, show, show_output,
                        sweep_id)
        run.launch()
        start_event = run.next_event()
        assert(start_event['type'] == 'start')
        if 'error' in start_event:
            raise RunnerError(start_event['error'])
        run_id = start_event['run_id']
        self._runs[run_id] = run
        return run_id

    def stop(self, run_id):
        print('Stop: %s' % run_id)
        if run_id in self._runs:
            self._runs[run_id].term()
        else:
            print('Run %s not running' % run_id)

    def running_runs(self):
        remove_runs = []
        for run_id, run in self._runs.items():
            if not run.is_running():
                remove_runs.append(run_id)
        for run in remove_runs:
            self._runs.pop(run)
        return list(self._runs.keys())



import multiprocessing
import socket
import sys
import traceback
from wandb.api import Api
from wandb import util
from wandb import runner


class AgentError(Exception):
    pass


class Agent(object):
    POLL_INTERVAL = 5

    def __init__(self, api, queue, wandb_runner, sweep_id=None):
        self._api = api
        self._queue = queue
        self._runner = wandb_runner
        self._server_responses = []
        self._sweep_id = sweep_id
        self._log = []

    def run(self):
        # TODO: include sweep ID
        agent = self._api.register_agent(
            socket.gethostname(), True, sweep_id=self._sweep_id)
        agent_id = agent['id']
        while True:
            commands = util.read_many_from_queue(
                self._queue, 100, self.POLL_INTERVAL)
            # TODO: send _server_responses
            running_runs = self._runner.running_runs()
            print('Running runs: ', running_runs)
            run_status = {}
            for run in running_runs:
                run_status[run] = True
            heartbeat_commands = self._api.agent_heartbeat(
                agent_id, {}, run_status)
            self._server_responses = []
            for command in heartbeat_commands:
                command['origin'] = 'server'
            commands += heartbeat_commands
            for command in commands:
                response = self._process_command(command)
                if command['origin'] == 'server':
                    self._server_responses.append(response)
                elif command['origin'] == 'local':
                    command['resp_queue'].put(response)

    def _process_command(self, command):
        print('Agent received command: %s' % command)
        response = {
            'id': command.get('id'),
            'result': None,
        }
        try:
            command_type = command['type']
            result = None
            if command_type == 'run':
                result = self._command_run(command)
            elif command_type == 'stop':
                result = self._command_stop(command)
            else:
                raise AgentError('No such command: %s' % command_type)
            response['result'] = result
        except:
            ex_type, ex, tb = sys.exc_info()
            response['exception'] = '%s: %s' % (ex_type.__name__, str(ex))
            response['traceback'] = traceback.format_tb(tb)
            del tb

        self._log.append((command, response))

        return response

    def _command_run(self, command):
        return self._runner.run(command['program'], command['args'],
                                sweep_id=self._sweep_id)

    def _command_stop(self, command):
        return self._runner.stop(command['run_id'])


class AgentApi(object):
    def __init__(self, queue):
        self._queue = queue
        self._command_id = 0
        self._multiproc_manager = multiprocessing.Manager()

    def command(self, command):
        command['origin'] = 'local'
        command['id'] = 'local-%s' % self._command_id
        self._command_id += 1
        resp_queue = self._multiproc_manager.Queue()
        command['resp_queue'] = resp_queue
        self._queue.put(command)
        result = resp_queue.get()
        print('result:', result)
        if 'exception' in result:
            print('Exception occurred while running command')
            for line in result['traceback']:
                print(line.strip())
            print(result['exception'])
        return result


def run_agent(sweep_id=None):
    api = Api()
    wandb_runner = runner.Runner(api)
    queue = multiprocessing.Queue()
    agent = Agent(api, queue, wandb_runner, sweep_id=sweep_id)
    p = multiprocessing.Process(target=agent.run)
    p.start()
    return AgentApi(queue)
